'''
@author: saaj
'''


import sys
import unittest
import types
import hashlib
import random
import datetime

try:
  import cPickle as pickle
except ImportError:
  import pickle

import hermes.backend


class TestCase(unittest.TestCase):

  testee = None


  def testMethodCoverage(self):
    if self.__class__ is TestCase:
      return

    def methods(object):
      return {
        name for (name, value) in object.__class__.__dict__.items()
        if isinstance(value, types.FunctionType) and name[0] != '_'
      }

    self.assertFalse(self.testee is None, 'Testee must be created in setUp()')

    diff = set('test' + name[0].upper() + name[1:] for name in methods(self.testee)) - methods(self)
    self.assertEqual(0, len(diff), 'Test case misses: {0}'.format(', '.join(diff)))
    
  def _arghash(self, *args, **kwargs):
    '''Not very neat as it penetrates into an implementation detail, though otherwise it'll be
    harder to make assertion on keys, because pickled results are different on py2 and py3'''
    
    arguments = args, tuple(sorted(kwargs.items()))
    return hashlib.md5(pickle.dumps(arguments, protocol = pickle.HIGHEST_PROTOCOL)).hexdigest()[::2]


def createFixture(cache):
  
  class Fixture:
  
    calls = 0
    
    
    @cache
    def simple(self, a, b):
      '''Here be dragons... seriously just a docstring test'''
      
      self.calls += 1
      return '{0}+{1}'.format(a, b)[::-1]
    
    @cache
    def nested(self, a, b):
      self.calls += 1
      return self.simple(b, a)[::-1]
    
    @cache(tags = ('rock', 'tree'))
    def tagged(self, a, b):
      self.calls += 1
      return '{0}-{1}'.format(a, b)[::-2]
    
    @cache(tags = ('rock', 'ice'))
    def tagged2(self, a, b):
      self.calls += 1
      return '{0}%{1}'.format(a, b)[::-2]
    
    @cache(tags = ('ash', 'stone'), key = lambda fn, *args, **kwargs: 'mykey:{0}:{1}'.format(*args))
    def key(self, a, b):
      self.calls += 1
      return '{0}*{1}'.format(a, b)[::2]
    
    @cache(tags = ('a', 'z'), key = lambda fn, *a: 'mk:{0}:{1}'.format(*a).replace(' ', ''), ttl = 1200)
    def all(self, a, b):
      self.calls += 1
      return {'a' : a['alpha'], 'b' : {'b' : b[0]}}
    
  return Fixture()


def benchmark(fixture):
  try:
    # Python 2
    xrange = xrange  # @UndefinedVariable
  except NameError:
    # Python 3
    xrange = range
  
  args   = [(random.randint(0, 4096), random.randint(0, 4096)) for _ in xrange(1024)]
  repeat = 8192
  for method in ('simple', 'tagged'):
    print('\n{0} {1} {0}'.format('-' * 20, method))
    start = datetime.datetime.now()
    for _ in xrange(repeat):
      getattr(fixture, method)(*random.choice(args))
    delta = datetime.datetime.now() - start
    print('  repeated:    {0:,}'.format(repeat))
    print('  time:        {0}'.format(delta))
    print('  call/second: {0:,.0f}'.format(repeat / delta.total_seconds()))
  

class TestReadme(unittest.TestCase):
  
  def testUsage(self):
    import hermes.backend.redis

  
    cache = hermes.Hermes(hermes.backend.redis.Backend, ttl = 600, host = 'localhost', db = 1)
        
    @cache
    def foo(a, b):
      return a * b
    
    class Example:
          
      @cache(tags = ('math', 'power'), ttl = 1200)
      def bar(self, a, b):
        return a ** b
        
      @cache(tags = ('math', 'avg'), key = lambda fn, *args, **kwargs: 'avg:{0}:{1}'.format(*args))
      def baz(self, a, b):
        return (a + b) / 2.0

          
    print(foo(2, 333))
    
    example = Example()
    print(example.bar(2, 10))
    print(example.baz(2, 10))
        
    foo.invalidate(2, 333)
    example.bar.invalidate(2, 10)
    example.baz.invalidate(2, 10)
        
    cache.clean(['math']) # invalidate entries tagged 'math'
    cache.clean()         # flush cache
  
  def testNonTagged(self):
    import hermes.backend.dict
  
    cache = hermes.Hermes(hermes.backend.dict.Backend)
        
    @cache
    def foo(a, b):
      return a * b
    
    foo(2, 2)
    foo(2, 4)
    
    print(cache.backend.dump()) 
    #  {
    #    'cache:entry:foo:515d5cb1a98de31d': 8, 
    #    'cache:entry:foo:a1c97600eac6febb': 4
    #  }
    
  def testTagged(self):
    import hermes.backend.dict
  
    cache = hermes.Hermes(hermes.backend.dict.Backend)
        
    @cache(tags = ('tag1', 'tag2'))
    def foo(a, b):
      return a * b
    
    foo(2, 2)
    
    print(cache.backend.dump()) 
    #  {
    #    u'cache:tag:tag1': '0674536f9eb4eb19', 
    #    u'cache:tag:tag2': 'db22b5ab2e504895', 
    #    'cache:entry:foo:a1c97600eac6febb:c1da510b3d42bad6': 4
    #  }
    
  def testTradeoff(self):
    import hermes.backend.dict
  
    cache = hermes.Hermes(hermes.backend.dict.Backend)
        
    @cache(tags = ('tag1', 'tag2'))
    def foo(a, b):
      return a * b
    
    foo(2, 2)
    
    print(cache.backend.dump())
    #  {
    #    u'cache:tag:tag1': '047820ac777abe8a', 
    #    u'cache:tag:tag2': '126365ec7175e851', 
    #    'cache:entry:foo:a1c97600eac6febb:5cae80f5e7d58329': 4
    #  }
    
    cache.clean(['tag1'])
    foo(2, 2)
    
    print(cache.backend.dump()) 
    #  {
    #    u'cache:tag:tag1': '66336fec212def16', 
    #    u'cache:tag:tag2': '126365ec7175e851', 
    #    'cache:entry:foo:a1c97600eac6febb:8e7e24cf70c1f0ab': 4, 
    #    'cache:entry:foo:a1c97600eac6febb:5cae80f5e7d58329': 4
    #  }


class TestWrapping(unittest.TestCase):
  
  testee = None
  
  
  def setUp(self):
    self.testee = hermes.Hermes() 
  
  def testFunction(self):
    
    @self.testee
    def foo(a, b):
      '''Overwhelmed everyone would be...'''
      
      return a * b
    
    self.assertTrue(isinstance(foo, hermes.Cached))
    self.assertEqual('foo', foo.__name__)
    self.assertEqual('Overwhelmed everyone would be...', foo.__doc__)
  
  def testMethod(self):
    fixture = createFixture(self.testee)
    
    self.assertTrue(isinstance(fixture.simple, hermes.Cached))
    self.assertEqual('simple', fixture.simple.__name__)
    self.assertEqual('Here be dragons... seriously just a docstring test', fixture.simple.__doc__)
    
  def testInstanceIsolation(self):
    
    class Fixture(object):
      
      def __init__(self, marker):
        self.marker = marker
      
      @self.testee  
      def foo(self):
        return self.marker
      
      def bar(self):
        pass
      
    f1 = Fixture(12)
    f2 = Fixture(24)
    
    # verify instances are not shared
    self.assertIsNot(f1.foo, f2.foo)
    # like it is normally the case of normal methods
    self.assertIsNot(f1.bar, f2.bar)
    
    self.assertEqual(12, f1.foo())
    self.assertEqual(24, f2.foo())




'''
@author: saaj
'''


import threading
import time
import json

import hermes.test as test
import hermes.backend.dict


class TestDict(test.TestCase):
  
  def setUp(self):
    self.testee  = hermes.Hermes(hermes.backend.dict.Backend, ttl = 360) 
    self.fixture = test.createFixture(self.testee) 
    
    self.testee.clean()
  
  def testSimple(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', b = 'beta'))
      self.assertEqual(1, self.fixture.calls)
      self.assertEqual({
        'cache:entry:Fixture:simple:' + self._arghash('alpha', b = 'beta') : 'ateb+ahpla'
      }, self.testee.backend.dump())
    
    self.fixture.simple.invalidate('alpha', b = 'beta')
    self.assertEqual({}, self.testee.backend.dump())

    
    self.assertEqual(1,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    expectedPy2 = "])]'ammag'[(tes[+}]'ateb'[ :'ahpla'{"
    expectedPy3 = "]}'ammag'{[+}]'ateb'[ :'ahpla'{"
    for _ in range(4):
      try:
        self.assertEqual(expectedPy2, self.fixture.simple({'alpha' : ['beta']}, [{'gamma'}]))
      except AssertionError:
        self.assertEqual(expectedPy3, self.fixture.simple({'alpha' : ['beta']}, [{'gamma'}]))
        
      self.assertEqual(2, self.fixture.calls)
      expected = {
        'cache:entry:Fixture:simple:' + self._arghash({'alpha' : ['beta']}, [{'gamma'}]) : expectedPy2
      }
      try:
        self.assertEqual(expected, self.testee.backend.dump())
      except AssertionError:
        expected[tuple(expected.keys())[0]] = expectedPy3
        self.assertEqual(expected, self.testee.backend.dump())
        
    
    self.fixture.simple.invalidate({'alpha' : ['beta']}, [{'gamma'}])
    self.assertEqual({}, self.testee.backend.dump())
    
  def testTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', b = 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      
      cache = self.testee.backend.dump()
      self.assertEqual(3, len(cache))
      self.assertEqual(16, len(cache.pop('cache:tag:rock')))
      self.assertEqual(16, len(cache.pop('cache:tag:tree')))
      
      expected = 'cache:entry:Fixture:tagged:' + self._arghash('alpha', b = 'beta')
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('ae-hl', tuple(cache.values())[0])
    
    
    self.fixture.tagged.invalidate('alpha', b = 'beta')
    
    cache = self.testee.backend.dump()
    self.assertEqual(2, len(cache))
    rockTag = cache.get('cache:tag:rock')
    treeTag = cache.get('cache:tag:tree')
    self.assertNotEqual(rockTag, treeTag)
    self.assertEqual(16, len(rockTag))
    self.assertEqual(16, len(treeTag))
    
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(3, self.fixture.calls)
      
      cache = self.testee.backend.dump()
      self.assertEqual(5, len(cache))
      self.assertEqual(rockTag, cache.get('cache:tag:rock'))
      self.assertEqual(treeTag, cache.get('cache:tag:tree'))
      self.assertEqual(16, len(cache.get('cache:tag:ice')))
      
    self.testee.clean(['rock'])
    
    cache = self.testee.backend.dump()
    self.assertEqual(4, len(cache))
    self.assertTrue('cache:tag:rock' not in cache)
    iceTag = cache.get('cache:tag:ice')
    
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(5, self.fixture.calls)
      
      cache = self.testee.backend.dump()
      self.assertEqual(7,  len(cache), 'has new and old entries for tagged and tagged 2 + 3 tags')
      self.assertEqual(treeTag, cache.get('cache:tag:tree'))
      self.assertEqual(iceTag,  cache.get('cache:tag:ice'))
      self.assertEqual(16, len(cache.get('cache:tag:rock')))
      self.assertNotEqual(rockTag, cache.get('cache:tag:rock'))
    
  def testFunction(self):
    counter = dict(foo = 0, bar = 0)
    
    @self.testee
    def foo(a, b):
      counter['foo'] += 1
      return '{0}+{1}'.format(a, b)[::-1]
    
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)
    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      counter['bar'] += 1
      return '{0}-{1}'.format(a, b)[::2]
    
    
    self.assertEqual(0,  counter['foo'])
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual('ateb+ahpla', foo('alpha', 'beta'))
      self.assertEqual(1, counter['foo'])
      self.assertEqual(
        {'cache:entry:foo:' + self._arghash('alpha', 'beta') : 'ateb+ahpla'}, 
        self.testee.backend.dump()
      )
    
    foo.invalidate('alpha', 'beta')
    self.assertEqual(1,  counter['foo'])
    self.assertEqual({}, self.testee.backend.dump())


    self.assertEqual(0,  counter['bar'])
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual('apabt', bar('alpha', 'beta'))
      self.assertEqual(1,       counter['bar'])
      
      cache = self.testee.backend.dump()
      self.assertEqual(3,  len(cache))
      self.assertEqual(16, len(cache.pop('cache:tag:a')))
      self.assertEqual(16, len(cache.pop('cache:tag:z')))
      
      expected = 'mk:alpha:beta'
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('apabt', tuple(cache.values())[0])
    
    bar.invalidate('alpha', 'beta')
    self.assertEqual(1, counter['foo'])
    
    cache = self.testee.backend.dump()
    self.assertEqual(2,  len(cache))
    self.assertEqual(16, len(cache.pop('cache:tag:a')))
    self.assertEqual(16, len(cache.pop('cache:tag:z')))

  def testKey(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual('apabt', self.fixture.key('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      
      cache = self.testee.backend.dump()
      self.assertEqual(3,  len(cache))
      self.assertEqual(16, len(cache.pop('cache:tag:ash')))
      self.assertEqual(16, len(cache.pop('cache:tag:stone')))
      
      expected = 'mykey:alpha:beta'
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('apabt', tuple(cache.values())[0])
    
    self.fixture.key.invalidate('alpha', 'beta')
    
    cache = self.testee.backend.dump()
    self.assertEqual(2,  len(cache))
    self.assertEqual(16, len(cache.pop('cache:tag:ash')))
    self.assertEqual(16, len(cache.pop('cache:tag:stone')))
    
  def testAll(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, self.fixture.all({'alpha' : 1}, ['beta']))
      self.assertEqual(1, self.fixture.calls)
      
      cache = self.testee.backend.dump()
      self.assertEqual(3, len(cache))
      self.assertEqual(16, len(cache.pop('cache:tag:a')))
      self.assertEqual(16, len(cache.pop('cache:tag:z')))
      
      expected = "mk:{'alpha':1}:['beta']"
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, tuple(cache.values())[0])
    
    self.fixture.all.invalidate({'alpha' : 1}, ['beta'])
    
    cache = self.testee.backend.dump()
    self.assertEqual(2,  len(cache))
    self.assertEqual(16, len(cache.pop('cache:tag:a')))
    self.assertEqual(16, len(cache.pop('cache:tag:z')))
  
  def testClean(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    
    self.testee.clean()
    
    self.assertEqual(2,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
  def testCleanTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    
    cache = self.testee.backend.dump()
    self.assertEqual(4, len(cache))
    
    self.assertEqual('ateb+ahpla', cache.pop('cache:entry:Fixture:simple:' + self._arghash('alpha', 'beta')))
    self.assertEqual(16, len(cache.pop('cache:tag:rock')))
    self.assertEqual(16, len(cache.pop('cache:tag:tree')))
    
    expectedKey = 'cache:entry:Fixture:tagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])
    
    
    self.testee.clean(('rock',))
    
    cache = self.testee.backend.dump()
    self.assertEqual(3, len(cache))
    self.assertEqual('ateb+ahpla', cache.pop('cache:entry:Fixture:simple:' + self._arghash('alpha', 'beta')))
    self.assertEqual(16, len(cache.pop('cache:tag:tree')))
    self.assertFalse('cache:tag:rock' in cache)
    
    expectedKey = 'cache:entry:Fixture:tagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])

    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(3,            self.fixture.calls)
    
    cache = self.testee.backend.dump()
    self.assertEqual(5, len(cache), '+1 old tagged entry')
    self.assertEqual('ateb+ahpla', cache.pop('cache:entry:Fixture:simple:' + self._arghash('alpha', 'beta')))
    self.assertEqual(16, len(cache.pop('cache:tag:rock')))
    self.assertEqual(16, len(cache.pop('cache:tag:tree')))
    
    expectedKey = 'cache:entry:Fixture:tagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])
    
    
    self.testee.clean(('rock', 'tree'))
    
    cache = self.testee.backend.dump()
    self.assertEqual(3, len(cache))
    self.assertEqual('ateb+ahpla', cache.pop('cache:entry:Fixture:simple:' + self._arghash('alpha', 'beta')))
    self.assertFalse('cache:tag:tree' in cache)
    self.assertFalse('cache:tag:rock' in cache)
    
    expectedKey = 'cache:entry:Fixture:tagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])
    
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(4,            self.fixture.calls)
    
    cache = self.testee.backend.dump()
    self.assertEqual(6, len(cache), '+2 old tagged entries')
    self.assertEqual('ateb+ahpla', cache.pop('cache:entry:Fixture:simple:' + self._arghash('alpha', 'beta')))
    expectedKey = 'cache:entry:Fixture:tagged:' + self._arghash('gamma', 'delta')
    self.assertTrue(expectedKey in (':'.join(k.split(':')[:-1]) for k in cache.keys()))
    self.assertTrue('aldamg' in cache.values())
    
    
    self.testee.clean()
    
    self.assertEqual(4,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
  
  def testNested(self):
    self.assertEqual('beta+alpha', self.fixture.nested('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)
    self.assertEqual({
      'cache:entry:Fixture:nested:' + self._arghash('alpha', 'beta') : 'beta+alpha', 
      'cache:entry:Fixture:simple:' + self._arghash('beta', 'alpha') : 'ahpla+ateb'
    }, self.testee.backend.dump())
    
  def testConcurrent(self):
    log = []
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)
    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      log.append(1)
      time.sleep(0.04)
      return '{0}-{1}'.format(a, b)[::2]
    
    threads = tuple(map(lambda i: threading.Thread(target = bar, args = ('alpha', 'beta')), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertEqual(1, sum(log))
    
    cache = self.testee.backend.dump()
    self.assertEqual(16, len(cache.pop('cache:tag:a')))
    self.assertEqual(16, len(cache.pop('cache:tag:z')))
    
    self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('apabt', tuple(cache.values())[0])
    
    del log[:]
    self.testee.clean()
    self.testee.backend.lock = lambda k: hermes.backend.AbstractLock(k) # now see a dogpile 
    
    threads = tuple(map(lambda i: threading.Thread(target = bar, args = ('alpha', 'beta')), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertGreater(sum(log), 1, 'dogpile')
    
    cache = self.testee.backend.dump()
    self.assertEqual(16, len(cache.pop('cache:tag:a')))
    self.assertEqual(16, len(cache.pop('cache:tag:z')))
    
    self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('apabt', tuple(cache.values())[0])


class TestDictLock(test.TestCase):
  
  def setUp(self):
    self.testee = hermes.backend.dict.Lock('123')
  
  def testAcquire(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertTrue(self.testee.acquire(False)) # reintrant within one thread
      finally:
        self.testee.release()
    
  def testRelease(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertTrue(self.testee.acquire(False)) # reintrant within one thread
      finally:
        self.testee.release()
    
  def testWith(self):
    with self.testee:
      self.assertTrue(self.testee.acquire(False)) # reintrant within one thread
      
  def testConcurrent(self):
    log   = []
    check = threading.Lock()
    def target():
      with self.testee:
        log.append(check.acquire(False))
        time.sleep(0.05)
        check.release()
        time.sleep(0.05)
    
    threads = tuple(map(lambda i: threading.Thread(target = target), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertEqual([True] * 4, log)


class CustomMangler(hermes.Mangler):
  
  prefix = 'hermes'
  '''Prefix for cache and tag entries'''
  

  def hash(self, value):
    return str(hash(value))
  
  def dumps(self, value):
    return json.dumps(value)
  
  def loads(self, value):
    return json.loads(value)
  

class TestDictCustomMangler(TestDict):
  
  def setUp(self):
    self.testee  = hermes.Hermes(hermes.backend.dict.Backend, CustomMangler, ttl = 360) 
    self.fixture = test.createFixture(self.testee) 
    
    self.testee.clean()
  
  def _arghash(self, *args, **kwargs):
    '''Not very neat as it penetrates into an implementation detail, though otherwise it'll 
    be harder to make assertion on keys, because hash results are different on py2 and py3
    and their 32/64-bit builds'''
    
    arguments = args, tuple(sorted(kwargs.items()))
    return hash(json.dumps(arguments))
  
  def testSimple(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
      self.assertEqual(1, self.fixture.calls)
      self.assertEqual({
        'hermes:entry:Fixture:simple:' + str(self._arghash('alpha', 'beta')) : 'ateb+ahpla'
      }, self.testee.backend.dump())
    
    self.fixture.simple.invalidate('alpha', 'beta')
    self.assertEqual({}, self.testee.backend.dump())

    
    self.assertEqual(1,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    expected = "]}'atled' :'ammag'{[+}]'ateb'[ :'ahpla'{"
    for _ in range(4):
      self.assertEqual(expected, self.fixture.simple({'alpha' : ['beta']}, [{'gamma' : 'delta'}]))
      self.assertEqual(2, self.fixture.calls)
      argHash = str(self._arghash({'alpha' : ['beta']}, [{'gamma' : 'delta'}]))
      self.assertEqual({'hermes:entry:Fixture:simple:' + argHash : expected}, self.testee.backend.dump())
    
    self.fixture.simple.invalidate({'alpha' : ['beta']}, [{'gamma' : 'delta'}])
    self.assertEqual({}, self.testee.backend.dump())
    
  def testTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      
      cache = self.testee.backend.dump()
      self.assertTrue(3, len(cache))
      self.assertFalse(cache.get('hermes:tag:tree') == cache.get('hermes:tag:rock'))
      self.assertTrue(int(cache.pop('hermes:tag:rock')) != 0)
      self.assertTrue(int(cache.pop('hermes:tag:tree')) != 0)
      
      expected = 'hermes:entry:Fixture:tagged:' + str(self._arghash('alpha', 'beta'))
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('ae-hl', tuple(cache.values())[0])
    
    self.fixture.tagged.invalidate('alpha', 'beta')
    
    cache = self.testee.backend.dump()
    self.assertEqual(2, len(cache))
    rockTag = cache.get('hermes:tag:rock')
    treeTag = cache.get('hermes:tag:tree')
    self.assertNotEqual(rockTag, treeTag)
    self.assertTrue(int(rockTag) != 0)
    self.assertTrue(int(treeTag) != 0)
    
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(3, self.fixture.calls)
      
      cache = self.testee.backend.dump()
      self.assertEqual(5,  len(cache))
      self.assertEqual(rockTag, cache.get('hermes:tag:rock'))
      self.assertEqual(treeTag, cache.get('hermes:tag:tree'))
      self.assertTrue(int(cache.get('hermes:tag:ice')) != 0)
      
    self.testee.clean(['rock'])
    
    cache = self.testee.backend.dump()
    self.assertEqual(4, len(cache))
    self.assertTrue('hermes:tag:rock' not in cache)
    iceTag = cache.get('hermes:tag:ice')
    
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(5, self.fixture.calls)
      
      cache = self.testee.backend.dump()
      self.assertEqual(7,  len(cache), 'has new and old entries for tagged and tagged 2 + 3 tags')
      self.assertEqual(treeTag, cache.get('hermes:tag:tree'))
      self.assertEqual(iceTag,  cache.get('hermes:tag:ice'))
      self.assertTrue(int(cache.get('hermes:tag:rock')) != 0)
      self.assertNotEqual(rockTag, cache.get('hermes:tag:rock'))
    
  def testFunction(self):
    counter = dict(foo = 0, bar = 0)
    
    @self.testee
    def foo(a, b):
      counter['foo'] += 1
      return '{0}+{1}'.format(a, b)[::-1]
    
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)
    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      counter['bar'] += 1
      return '{0}-{1}'.format(a, b)[::2]
    
    
    self.assertEqual(0,  counter['foo'])
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual('ateb+ahpla', foo('alpha', 'beta'))
      self.assertEqual(1, counter['foo'])
      argHash = self._arghash('alpha', 'beta')
      self.assertEqual({'hermes:entry:foo:' + str(argHash) : 'ateb+ahpla'}, self.testee.backend.dump())
    
    foo.invalidate('alpha', 'beta')
    self.assertEqual(1,  counter['foo'])
    self.assertEqual({}, self.testee.backend.dump())


    self.assertEqual(0,  counter['bar'])
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual('apabt', bar('alpha', 'beta'))
      self.assertEqual(1,       counter['bar'])
      
      cache = self.testee.backend.dump()
      self.assertTrue(cache.get('hermes:tag:a') != cache.get('hermes:tag:z'))
      self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
      self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)
      
      self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('apabt', tuple(cache.values())[0])
    
    bar.invalidate('alpha', 'beta')
    
    
    self.assertEqual(1,  counter['foo'])
    
    cache = self.testee.backend.dump()
    self.assertTrue(cache.get('hermes:tag:a') != cache.get('hermes:tag:z'))
    self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)

  def testKey(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual('apabt', self.fixture.key('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      
      cache = self.testee.backend.dump()
      self.assertTrue(cache.get('hermes:tag:ash') != cache.get('hermes:tag:stone'))
      self.assertTrue(int(cache.pop('hermes:tag:ash')) != 0)
      self.assertTrue(int(cache.pop('hermes:tag:stone')) != 0)
      
      self.assertEqual('mykey:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('apabt', tuple(cache.values())[0])
    
    self.fixture.key.invalidate('alpha', 'beta')
    
    cache = self.testee.backend.dump()
    self.assertTrue(cache.get('hermes:tag:ash') != cache.get('hermes:tag:stone'))
    self.assertTrue(int(cache.pop('hermes:tag:ash')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:stone')) != 0)
    
  def testAll(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    for _ in range(4):
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, self.fixture.all({'alpha' : 1}, ['beta']))
      self.assertEqual(1, self.fixture.calls)
      
      cache = self.testee.backend.dump()
      self.assertEqual(3, len(cache))
      self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
      self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)
      
      expected = "mk:{'alpha':1}:['beta']"
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, tuple(cache.values())[0])
    
    self.fixture.all.invalidate({'alpha' : 1}, ['beta'])
    
    cache = self.testee.backend.dump()
    self.assertEqual(2, len(cache))
    self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)
  
  def testClean(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    
    self.testee.clean()
    
    self.assertEqual(2,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
  def testCleanTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    
    cache = self.testee.backend.dump()
    self.assertTrue(4, len(cache))
    self.assertEqual('ateb+ahpla', cache.pop('hermes:entry:Fixture:simple:' + str(self._arghash('alpha', 'beta'))))
    self.assertFalse(cache.get('hermes:tag:tree') == cache.get('hermes:tag:rock'))
    self.assertTrue(int(cache.pop('hermes:tag:rock')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:tree')) != 0)
    
    expected = 'hermes:entry:Fixture:tagged:' + str(self._arghash('gamma', 'delta'))
    self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])
    
    
    self.testee.clean(('rock',))
    
    cache = self.testee.backend.dump()
    self.assertTrue(3, len(cache))
    self.assertEqual('ateb+ahpla', cache.pop('hermes:entry:Fixture:simple:' + str(self._arghash('alpha', 'beta'))))
    self.assertTrue(int(cache.pop('hermes:tag:tree')) != 0)
    self.assertFalse('hermes:tag:rock' in cache)
    
    self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])

    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(3,            self.fixture.calls)
    
    cache = self.testee.backend.dump()
    self.assertTrue(4, len(cache))
    self.assertEqual('ateb+ahpla', cache.pop('hermes:entry:Fixture:simple:' + str(self._arghash('alpha', 'beta'))))
    self.assertFalse(cache.get('hermes:tag:tree') == cache.get('hermes:tag:rock'))
    self.assertTrue(int(cache.pop('hermes:tag:rock')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:tree')) != 0)
    
    self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])
    
    
    self.testee.clean(('rock', 'tree'))
    
    cache = self.testee.backend.dump()
    self.assertTrue(2, len(cache))
    self.assertEqual('ateb+ahpla', cache.pop('hermes:entry:Fixture:simple:' + str(self._arghash('alpha', 'beta'))))
    self.assertFalse('hermes:tag:tree' in cache)
    self.assertFalse('hermes:tag:rock' in cache)
    
    self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])
    
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(4,            self.fixture.calls)
    
    cache = self.testee.backend.dump()
    self.assertTrue(4, len(cache))
    self.assertEqual('ateb+ahpla', cache.pop('hermes:entry:Fixture:simple:' + str(self._arghash('alpha', 'beta'))))
    self.assertFalse(cache.get('hermes:tag:tree') == cache.get('hermes:tag:rock'))
    self.assertTrue(int(cache.pop('hermes:tag:rock')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:tree')) != 0)
    
    self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])
    
    
    self.testee.clean()
    
    self.assertEqual(4,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())
  
  def testNested(self):
    self.assertEqual('beta+alpha', self.fixture.nested('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)
    self.assertEqual({
      'hermes:entry:Fixture:nested:' + str(self._arghash('alpha', 'beta')) : 'beta+alpha', 
      'hermes:entry:Fixture:simple:' + str(self._arghash('beta', 'alpha')) : 'ahpla+ateb'
    }, self.testee.backend.dump())
    
  def testConcurrent(self):
    log = []
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)
    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      log.append(1)
      time.sleep(0.04)
      return '{0}-{1}'.format(a, b)[::2]
    
    threads = tuple(map(lambda i: threading.Thread(target = bar, args = ('alpha', 'beta')), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertEqual(1, sum(log))
    
    cache = self.testee.backend.dump()
    self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)
    
    self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('apabt', tuple(cache.values())[0])
    
    del log[:]
    self.testee.clean()
    self.testee.backend.lock = lambda k: hermes.backend.AbstractLock(k) # now see a dogpile 
    
    threads = tuple(map(lambda i: threading.Thread(target = bar, args = ('alpha', 'beta')), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertGreater(sum(log), 1, 'dogpile')
    
    cache = self.testee.backend.dump()
    self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)
    
    self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('apabt', tuple(cache.values())[0])


class TestDictPerformance(test.unittest.TestCase):
  
  def setUp(self):
    self.testee  = hermes.Hermes(hermes.backend.dict.Backend, ttl = 360) 
    self.fixture = test.createFixture(self.testee)
    
  def tearDown(self):
    self.testee.clean()
  
  def testPerformance(self):
    test.benchmark(self.fixture)




'''
@author: saaj
'''


import collections


class AbstractLock(object):
  '''Base locking class. Implements context manger protocol. Mocks ``acquire`` and ``release``
  i.e. it always acquires.'''
  
  key = None
  '''Implementation may be key-aware'''


  def __init__(self, key):
    self.key = key

  def __enter__(self):
    self.acquire()

  def __exit__(self, type, value, traceback):
    self.release()
    
  def acquire(self, wait = True):
    return True

  def release(self):
    pass


class AbstractBackend(object):
  '''Abstract backend'''

  mangler = None
  '''Key manager responsible for creating keys, hashing and serialzation'''
  
  try:
    __str = basestring  # @UndefinedVariable
  except NameError:
    __str = (bytes, str)
  '''Python 2/3 compatible base string type'''


  def __init__(self, mangler, **kwargs):
    self.mangler = mangler
    
  @classmethod
  def _isScalar(cls, value):
    return not isinstance(value, collections.Iterable) or isinstance(value, cls.__str)

  def lock(self, key):
    return AbstractLock(self.mangler.nameLock(key))

  def save(self, key = None, value = None, mapping = None, ttl = None):
    pass
  
  def load(self, keys):
    '''Note, when handling a multiple key call, absent value keys 
    should be excluded from resulting dictionary''' 
    
    return None if self._isScalar(keys) else {} 
  
  def remove(self, keys):
    pass
  
  def clean(self):
    pass




'''
@author: saaj
'''


import threading

from . import AbstractBackend, AbstractLock


__all__ = 'Lock', 'Backend'


class Lock(AbstractLock):
  '''Key-unaware thread lock'''
  
    
  _lock = None
  '''Threading lock instance'''  
  
  def __init__(self, key):
    self._lock = threading.RLock()

  def acquire(self, wait = True):
    return self._lock.acquire(wait)

  def release(self):
    self._lock.release()


class Backend(AbstractBackend):
  '''Test purpose backend implementation. Does not support entry expiry. ``save`` and ``delete``
  are not atomic in general. Though because writes are synchronized it may be suitable for 
  limited number of real cases with small cached entry size.'''
  
  
  cache = None
  '''A dict intance'''
  
  _lock = None
  '''Lock instance'''
  
  
  def __init__(self, mangler, **kwargs):
    super(Backend, self).__init__(mangler)
    
    self.cache = {}
    self._lock = Lock(None)
  
  def lock(self, key):
    return self._lock
  
  def save(self, key = None, value = None, mapping = None, ttl = None):
    if not mapping:
      mapping = {key : value}

    self.cache.update({k : self.mangler.dumps(v) for k, v in mapping.items()})
  
  def load(self, keys):
    if self._isScalar(keys):
      value = self.cache.get(keys, None)
      if value is not None:
        value = self.mangler.loads(value)
      return value
    else:
      return {k : self.mangler.loads(self.cache[k]) for k in keys if k in self.cache}
  
  def remove(self, keys):
    if self._isScalar(keys):
      keys = (keys,)
      
    for key in keys:
      self.cache.pop(key, None)  

  def clean(self):
    self.cache.clear()
    
  def dump(self):
    return {k : self.mangler.loads(v) for k, v in self.cache.items()}




'''
@author: saaj
'''


import threading
import time

import hermes.test as test
import hermes.backend


class TestAbstract(test.TestCase):
  
  def setUp(self):
    self.testee  = hermes.Hermes(hermes.backend.AbstractBackend, ttl = 360) 
    self.fixture = test.createFixture(self.testee) 
    
    self.testee.clean()
  
  def testSimple(self):
    self.assertEqual(0, self.fixture.calls)
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', b = 'beta'))
    self.assertEqual(1, self.fixture.calls)
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)
    
  def testTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    
    self.assertEqual('ae-hl', self.fixture.tagged('alpha', b = 'beta'))
    self.assertEqual(1,       self.fixture.calls)
    
    self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
    self.assertEqual(2,       self.fixture.calls)
    
  def testFunction(self):
    counter = dict(foo = 0, bar = 0)
    
    @self.testee
    def foo(a, b):
      counter['foo'] += 1
      return '{0}+{1}'.format(a, b)[::-1]
    
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)
    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      counter['bar'] += 1
      return '{0}-{1}'.format(a, b)[::2]
    
    
    self.assertEqual(0, counter['foo'])
    
    self.assertEqual('ateb+ahpla', foo('alpha', 'beta'))
    self.assertEqual(1, counter['foo'])
    
    self.assertEqual('ateb+ahpla', foo('alpha', 'beta'))
    self.assertEqual(2, counter['foo'])

    self.testee.clean()
    self.assertEqual(0, counter['bar'])
    
    self.assertEqual('apabt', bar('alpha', 'beta'))
    self.assertEqual(1,       counter['bar'])
    
    self.assertEqual('apabt', bar('alpha', 'beta'))
    self.assertEqual(2,       counter['bar'])

  def testKey(self):
    self.assertEqual(0,  self.fixture.calls)
    
    self.assertEqual('apabt', self.fixture.key('alpha', 'beta'))
    self.assertEqual(1,       self.fixture.calls)
    
    self.assertEqual('apabt', self.fixture.key('alpha', 'beta'))
    self.assertEqual(2,       self.fixture.calls)
    
  def testAll(self):
    self.assertEqual(0, self.fixture.calls)
    
    self.assertEqual( {'a': ['beta'], 'b': {'b': 2}}, self.fixture.all({'alpha' : ['beta']}, [2]))
    self.assertEqual(1, self.fixture.calls)
    
    self.assertEqual( {'a': ['beta'], 'b': {'b': 2}}, self.fixture.all({'alpha' : ['beta']}, [2]))
    self.assertEqual(2, self.fixture.calls)
    
  def testClean(self):
    self.assertEqual(0, self.fixture.calls)
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    
    self.testee.clean()
    self.assertEqual(2, self.fixture.calls)
    
  def testCleanTagged(self):
    self.assertEqual(0, self.fixture.calls)
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    
    self.testee.clean(('rock',))
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(4,            self.fixture.calls)
    
    self.testee.clean(('rock', 'tree'))
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(6,            self.fixture.calls)
    
    self.testee.clean()
    
    self.assertEqual(6,  self.fixture.calls)
    
  def testNested(self):
    self.assertEqual('beta+alpha', self.fixture.nested('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)


class TestAbstractLock(test.TestCase):
  
  def setUp(self):
    self.testee = hermes.backend.AbstractLock('123')
  
  def testAcquire(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertTrue(self.testee.acquire(False))
      finally:
        self.testee.release()
    
  def testRelease(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertTrue(self.testee.acquire(False))
      finally:
        self.testee.release()
    
  def testWith(self):
    with self.testee:
      self.assertTrue(self.testee.acquire(False))
      
  def testConcurrent(self):
    log   = []
    check = threading.Lock()
    def target():
      with self.testee:
        locked = check.acquire(False)
        log.append(locked)
        time.sleep(0.05)
        if locked:
          check.release()
        time.sleep(0.05)
    
    threads = tuple(map(lambda i: threading.Thread(target = target), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertEqual([True, False, False, False], log)


class TestAbstractPerformance(test.unittest.TestCase):
  
  def setUp(self):
    self.testee  = hermes.Hermes(hermes.backend.AbstractBackend, ttl = 360) 
    self.fixture = test.createFixture(self.testee)
    
  def tearDown(self):
    self.testee.clean()
  
  def testPerformance(self):
    test.benchmark(self.fixture)




'''
@author: saaj
'''


import types
import hashlib
import functools
import os
import binascii

try:
  import cPickle as pickle
except ImportError:
  import pickle

from .backend import AbstractBackend


__all__ = 'Hermes', 'Mangler'


class Mangler(object):
  '''Key manager responsible for creating keys, hashing and serialization'''
  
  prefix = 'cache'
  '''Prefix for cache and tag entries'''
  

  def hash(self, value):
    return hashlib.md5(value).hexdigest()[::2] # full md5 seems too long
  
  def dumps(self, value):
    return pickle.dumps(value, protocol = pickle.HIGHEST_PROTOCOL)
  
  def loads(self, value):
    return pickle.loads(value)
  
  def nameEntry(self, fn, *args, **kwargs):
    result = [self.prefix, 'entry']
    if isinstance(fn, types.MethodType):
      result.extend([fn.__self__.__class__.__name__, fn.__name__])
    elif isinstance(fn, types.FunctionType):
      result.append(fn.__name__)
    else:
      raise TypeError('Fn is expected to be insance of MethodType or FunctionType')

    arguments = args, tuple(sorted(kwargs.items()))
    result.append(self.hash(self.dumps(arguments))) 
    
    return ':'.join(result)

  def nameTag(self, tag):
    return u':'.join([self.prefix, 'tag', tag]) 

  def mapTags(self, tagKeys):
    hash = binascii.hexlify(os.urandom(4)).decode('ascii')
    return {key : self.hash(u':'.join((key, hash)).encode('utf8')) for key in tagKeys}

  def hashTags(self, tagMap):
    values = tuple(zip(*sorted(tagMap.items())))[1] # sorted by key dict values
    return self.hash(u':'.join(values).encode('utf8'))
  
  def nameLock(self, entryKey):
    parts = entryKey.split(':')
    if parts[0] == self.prefix:
      entryKey = ':'.join(parts[2:]) 
    return ':'.join([self.prefix, 'lock', entryKey])


class Hermes(object):
  '''Cache facade. Usage: 
    
      import hermes.backend.redis
    
      cache = hermes.Hermes(hermes.backend.redis.Backend, ttl = 600, host = 'localhost', db = 1)
          
          
      @cache
      def foo(a, b):
        return a * b
      
      class Example:
            
        @cache(tags = ('math', 'power'), ttl = 1200)
        def bar(self, a, b):
          return a ** b
          
        @cache(tags = ('math', 'avg'), key = lambda fn, *args, **kwargs: 'avg:{0}:{1}'.format(*args))
        def baz(self, a, b):
          return (a + b) / 2.0
  
            
      print(foo(2, 333))
      
      example = Example()
      print(example.bar(2, 10))
      print(example.baz(2, 10))
          
      foo.invalidate(2, 333)
      example.bar.invalidate(2, 10)
      example.baz.invalidate(2, 10)
          
      cache.clean(['math']) # invalidate entries tagged 'math'
      cache.clean()         # flush cache
  ''' 
  
  backend = None
  '''Cache backend'''
  
  mangler = None
  '''Key manager responsible for creating keys, hashing and serialization'''
  
  ttl = 3600
  '''Default cache entry Time To Live'''
  
  
  def __init__(self, backendClass = AbstractBackend, manglerClass = Mangler, **kwargs):
    '''Initialises the cache decorator factory. 
      
    Positional arguments are backend class and mangler class. If omitted noop-backend
    and built-in mangler will be be used.
    
    Keyword arguments comprise of ``ttl`` and backend parameters.
    '''
    
    self.ttl = kwargs.pop('ttl', self.ttl)
    
    assert issubclass(manglerClass, Mangler)
    self.mangler = manglerClass()
    
    assert issubclass(backendClass, AbstractBackend)
    self.backend = backendClass(self.mangler, **kwargs)

  def __call__(self, *args, **kwargs):
    '''Decorator that caches method or function result. The following key arguments are optional:
    
      :key:   Lambda that provides custom key, otherwise ``Mangler.nameEntry`` is used.
      :ttl:   Seconds until entry expiration, otherwise instance default is used.
      :tags:  Cache entry tag list.
      
    ``@cache`` decoration is supported as well as 
    ``@cache(ttl = 7200, tags = ('tag1', 'tag2'), key = lambda fn, *args, **kwargs: 'mykey')``.
    '''
    
    if args and isinstance(args[0], (types.FunctionType, types.MethodType)):
      # @cache
      return Cached(self.backend, self.mangler, self.ttl, args[0])
    else:
      # @cache()
      return lambda fn: Cached(self.backend, self.mangler, kwargs.pop('ttl', self.ttl), fn, **kwargs)
    
  def clean(self, tags = None):
    '''If tags argument is omitted flushes all entries, otherwise removes provided tag entries'''
    
    if tags:
      self.backend.remove(map(self.mangler.nameTag, tags))
    else:
      self.backend.clean()


class Cached(object):
  '''A wrapper for cached function or method'''
  
  _backend = None
  '''Cache backend'''
  
  _mangler = None
  '''Key manager responsible for creating keys, hashing and serializing values'''
  
  _callable = None
  '''The decorated callable, stays ``types.FunctionType`` if a function is decorated, 
  otherwise it is transformed to ``types.MethodType`` on the instance clone by descriptor 
  protocol implementation.
  '''
  
  _ttl = None
  '''Cache entry Time To Live for decorated callable'''
  
  _keyFunc  = None
  '''Key creation function'''
  
  _tags = None
  '''Cache entry tags for decorated callable'''
  
  
  def __init__(self, backend, mangler, ttl, callable, **kwargs):
    self._backend  = backend
    self._mangler  = mangler
    self._callable = callable
    self._ttl      = ttl
    self._keyFunc  = kwargs.get('key', self._mangler.nameEntry)
    self._tags     = kwargs.get('tags', None)
    # preserve ``__name__``, ``__doc__``, etc
    functools.update_wrapper(self, callable)
  
  def _load(self, key):
    if self._tags:
      tagMap = self._backend.load(map(self._mangler.nameTag, self._tags))
      if len(tagMap) != len(self._tags):
        return None
      else:
        key += ':' + self._mangler.hashTags(tagMap)
      
    return self._backend.load(key)
  
  def _save(self, key, value):
    if self._tags:
      namedTags   = tuple(map(self._mangler.nameTag, self._tags))
      tagMap      = self._backend.load(namedTags)
      missingTags = set(namedTags) - set(tagMap.keys())
      if missingTags:
        missingTagMap = self._mangler.mapTags(missingTags)
        self._backend.save(mapping = missingTagMap, ttl = None)
        tagMap.update(missingTagMap)
        assert len(self._tags) == len(tagMap) 
        
      key += ':' + self._mangler.hashTags(tagMap)
      
    return self._backend.save(key, value, ttl = self._ttl)
  
  def _remove(self, key):
    if self._tags:
      tagMap = self._backend.load(map(self._mangler.nameTag, self._tags))
      if len(tagMap) != len(self._tags):
        return
      else:
        key += ':' + self._mangler.hashTags(tagMap)
      
    self._backend.remove(key)
  
  def invalidate(self, *args, **kwargs):
    self._remove(self._keyFunc(self._callable, *args, **kwargs))
  
  def __call__(self, *args, **kwargs):
    key   = self._keyFunc(self._callable, *args, **kwargs)
    value = self._load(key)
    if value is None:
      with self._backend.lock(key):
        # it's better to read twice than lock every read
        value = self._load(key)
        if value is None:
          value = self._callable(*args, **kwargs)
          self._save(key, value)
    return value
  
  def __get__(self, instance, type):
    '''Implements non-data descriptor protocol.
    
    The invocation happens only when instance method is decorated, so we can distinguish 
    between decorated ``types.MethodType`` and ``types.FunctionType``. Python class 
    declaration mechanics prevent a decorator from having awareness of the class type, as 
    the function is received by the decorator before it becomes an instance method.
    
    How it works::
     
      cache = hermes.Hermes()
      
      class Model:
      
        @cache
        def calc(self):
          return 42 
      
      m = Model()
      m.calc
      
    Last attribute access results in the call, ``calc.__get__(m, Model)``, where 
    ``calc`` is instance of ``hermes.Cached`` which decorates the original ``Model.calc``.
    
    Note, initially ``hermes.Cached`` is created on decoration per class method, when class 
    type is created by the interpreter, and is shared among all instances. Later, on attribute 
    access, a copy is returned with bound ``_callable``, just like ordinary Python method 
    descriptor works.    
    
    For more details, http://docs.python.org/2/howto/descriptor.html#descriptor-protocol.
    '''
    
    if not isinstance(self._callable, types.MethodType):
      methodCached          = object.__new__(Cached)
      methodCached.__dict__ = self.__dict__.copy()
      try:
        methodCached._callable = types.MethodType(self._callable, instance, type)
      except TypeError:
        # python3 compatibility
        methodCached._callable = types.MethodType(self._callable, instance)
        
      return methodCached
    
    return self




'''
@author: saaj
'''


import threading
import time
import pickle

import hermes.test as test
import hermes.backend.redis


class TestRedis(test.TestCase):
  
  def setUp(self):
    self.testee = hermes.Hermes(hermes.backend.redis.Backend, ttl = 360, lockTimeout = 120) 
    self.fixture = test.createFixture(self.testee) 
    
    self.testee.clean()
    
  def tearDown(self):
    self.testee.clean()
  
  def testSimple(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.testee.backend.client.dbsize())

    key = 'cache:entry:Fixture:simple:' + self._arghash('alpha', b = 'beta')
    for _ in range(4):    
      self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', b = 'beta'))
      self.assertEqual(1, self.fixture.calls)
      self.assertEqual(1, self.testee.backend.client.dbsize())
  
      self.assertEqual(360, self.testee.backend.client.ttl(key))
      self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(key)))
    
    self.fixture.simple.invalidate('alpha', b = 'beta')
    self.assertEqual(0, self.testee.backend.client.dbsize())
    
    
    expectedPy2 = "])]'ammag'[(tes[+}]'ateb'[ :'ahpla'{"
    expectedPy3 = "]}'ammag'{[+}]'ateb'[ :'ahpla'{"
    key         = 'cache:entry:Fixture:simple:' + self._arghash({'alpha' : ['beta']}, [{'gamma'}])
    for _ in range(4): 
      try:
        self.assertEqual(expectedPy2, self.fixture.simple({'alpha' : ['beta']}, [{'gamma'}]))
      except AssertionError:
        self.assertEqual(expectedPy3, self.fixture.simple({'alpha' : ['beta']}, [{'gamma'}]))
        
      self.assertEqual(2, self.fixture.calls)
      self.assertEqual(1, self.testee.backend.client.dbsize())
  
      self.assertEqual(360, self.testee.backend.client.ttl(key))
      
      try:
        self.assertEqual(expectedPy2, pickle.loads(self.testee.backend.client.get(key)))
      except AssertionError:
        self.assertEqual(expectedPy3, pickle.loads(self.testee.backend.client.get(key)))
    
  def testTagged(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.testee.backend.client.dbsize())
    
    for _ in range(4):    
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', b = 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      self.assertEqual(3,       self.testee.backend.client.dbsize())
      
      self.assertEqual(-1,  self.testee.backend.client.ttl('cache:tag:rock'))
      self.assertEqual(-1,  self.testee.backend.client.ttl('cache:tag:tree'))
      rockTag = pickle.loads(self.testee.backend.client.get('cache:tag:rock'))
      treeTag = pickle.loads(self.testee.backend.client.get('cache:tag:tree'))
      self.assertFalse(rockTag == treeTag)
      self.assertEqual(16, len(rockTag))
      self.assertEqual(16, len(treeTag))
      
      argHash = self._arghash('alpha', b = 'beta')
      tagHash = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
      key     = 'cache:entry:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
      self.assertEqual(360, self.testee.backend.client.ttl(key))
      self.assertEqual('ae-hl', pickle.loads(self.testee.backend.client.get(key)))
    
    self.fixture.tagged.invalidate('alpha', b = 'beta')
    
    self.assertEqual(2, self.testee.backend.client.dbsize())
    
    rockTag = pickle.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = pickle.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertNotEqual(rockTag, treeTag)
    self.assertEqual(16, len(rockTag))
    self.assertEqual(16, len(treeTag))
    
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(3, self.fixture.calls)
      
      self.assertEqual(5, self.testee.backend.client.dbsize())
      self.assertEqual(rockTag, pickle.loads(self.testee.backend.client.get('cache:tag:rock')))
      self.assertEqual(treeTag, pickle.loads(self.testee.backend.client.get('cache:tag:tree')))
      self.assertEqual(16, len(pickle.loads(self.testee.backend.client.get('cache:tag:ice'))))
      
    self.testee.clean(['rock'])
    
    self.assertEqual(4, self.testee.backend.client.dbsize())
    self.assertIsNone(self.testee.backend.client.get('cache:tag:rock'))
    iceTag = pickle.loads(self.testee.backend.client.get('cache:tag:ice'))
    
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(5, self.fixture.calls)
      
      size = self.testee.backend.client.dbsize()
      self.assertEqual(7, size, 'has new and old entries for tagged and tagged 2 + 3 tags')
      self.assertEqual(treeTag, pickle.loads(self.testee.backend.client.get('cache:tag:tree')))
      self.assertEqual(iceTag, pickle.loads(self.testee.backend.client.get('cache:tag:ice')))
      self.assertEqual(16, len(pickle.loads(self.testee.backend.client.get('cache:tag:rock'))))
      self.assertNotEqual(rockTag, pickle.loads(self.testee.backend.client.get('cache:tag:rock')))
    
  def testFunction(self):
    counter = dict(foo = 0, bar = 0)
    
    @self.testee
    def foo(a, b):
      counter['foo'] += 1
      return '{0}+{1}'.format(a, b)[::-1]
    
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)
    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      counter['bar'] += 1
      return '{0}-{1}'.format(a, b)[::2]
    
    
    self.assertEqual(0, counter['foo'])
    self.assertEqual(0, self.testee.backend.client.dbsize())
    
    key = 'cache:entry:foo:' + self._arghash('alpha', 'beta')
    for _ in range(4):
      self.assertEqual('ateb+ahpla', foo('alpha', 'beta'))
      
      self.assertEqual(1, counter['foo'])
      self.assertEqual(1, self.testee.backend.client.dbsize())

      self.assertEqual(360,          self.testee.backend.client.ttl(key))
      self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(key)))
    
    foo.invalidate('alpha', 'beta')
    self.assertFalse(self.testee.backend.client.exists(key))
    self.assertEqual(0, self.testee.backend.client.dbsize())
    
    
    self.assertEqual(0, counter['bar'])
    self.assertEqual(0, self.testee.backend.client.dbsize())
    
    for _ in range(4):
      self.assertEqual('apabt', bar('alpha', 'beta'))
      self.assertEqual(1,       counter['bar'])
      self.assertEqual(3,       self.testee.backend.client.dbsize())
      
      self.assertEqual(-1,  self.testee.backend.client.ttl('cache:tag:a'))
      self.assertEqual(-1,  self.testee.backend.client.ttl('cache:tag:z'))
      
      aTag = pickle.loads(self.testee.backend.client.get('cache:tag:a'))
      zTag = pickle.loads(self.testee.backend.client.get('cache:tag:z'))
      self.assertFalse(aTag == zTag)
      self.assertEqual(16, len(aTag))
      self.assertEqual(16, len(zTag))
      
      tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
      key     = 'mk:alpha:beta:' + tagHash
      self.assertEqual(120,     self.testee.backend.client.ttl(key))
      self.assertEqual('apabt', pickle.loads(self.testee.backend.client.get(key)))
    
    bar.invalidate('alpha', 'beta')
    self.assertEqual(1,  counter['foo'])
    self.assertIsNone(self.testee.backend.client.get(key))
    self.assertEqual(2, self.testee.backend.client.dbsize())
    
    self.assertEqual('apabt', bar('alpha', 'beta'))
    self.assertEqual(2, counter['bar'])
    self.assertEqual(3, self.testee.backend.client.dbsize())

  def testKey(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.testee.backend.client.dbsize())
    
    for _ in range(4):
      self.assertEqual('apabt', self.fixture.key('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      self.assertEqual(3,       self.testee.backend.client.dbsize())
      
      self.assertEqual(-1,  self.testee.backend.client.ttl('cache:tag:ash'))
      self.assertEqual(-1,  self.testee.backend.client.ttl('cache:tag:stone'))
      ashTag   = pickle.loads(self.testee.backend.client.get('cache:tag:ash'))
      stoneTag = pickle.loads(self.testee.backend.client.get('cache:tag:stone'))
      self.assertFalse(ashTag == stoneTag)
      self.assertEqual(16, len(ashTag))
      self.assertEqual(16, len(stoneTag))
      
      tagHash = self.testee.mangler.hashTags(dict(ash = ashTag, stone = stoneTag))
      key     = 'mykey:alpha:beta:' + tagHash
      self.assertEqual(360, self.testee.backend.client.ttl(key))
      self.assertEqual('apabt', pickle.loads(self.testee.backend.client.get(key)))
    
    self.fixture.key.invalidate('alpha', 'beta')
    
    self.assertIsNone(self.testee.backend.client.get(key))
    self.assertEqual(2, self.testee.backend.client.dbsize())
    
    self.assertEqual(ashTag,   pickle.loads(self.testee.backend.client.get('cache:tag:ash')))
    self.assertEqual(stoneTag, pickle.loads(self.testee.backend.client.get('cache:tag:stone')))
    
  def testAll(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.testee.backend.client.dbsize())
    for _ in range(4):
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, self.fixture.all({'alpha' : 1}, ['beta']))
      self.assertEqual(1, self.fixture.calls)
      self.assertEqual(3, self.testee.backend.client.dbsize())
      
      self.assertEqual(-1,   self.testee.backend.client.ttl('cache:tag:a'))
      self.assertEqual(-1,   self.testee.backend.client.ttl('cache:tag:z'))
      
      aTag = pickle.loads(self.testee.backend.client.get('cache:tag:a'))
      zTag = pickle.loads(self.testee.backend.client.get('cache:tag:z'))
      self.assertFalse(aTag == zTag)
      self.assertEqual(16, len(aTag))
      self.assertEqual(16, len(zTag))
      
      tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
      key = "mk:{'alpha':1}:['beta']:" + tagHash
      self.assertEqual(1200, self.testee.backend.client.ttl(key))
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, pickle.loads(self.testee.backend.client.get(key)))
      
    self.fixture.all.invalidate({'alpha' : 1}, ['beta'])
    
    self.assertEqual(2, self.testee.backend.client.dbsize())
    self.assertEqual(aTag, pickle.loads(self.testee.backend.client.get('cache:tag:a')))
    self.assertEqual(zTag, pickle.loads(self.testee.backend.client.get('cache:tag:z')))
  
  def testClean(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.testee.backend.client.dbsize())
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    self.assertEqual(4,            self.testee.backend.client.dbsize())
    
    self.testee.clean()
    
    self.assertEqual(2,  self.fixture.calls)
    self.assertEqual(0, self.testee.backend.client.dbsize())
    
  def testCleanTagged(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.testee.backend.client.dbsize())
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    self.assertEqual(4,            self.testee.backend.client.dbsize())

    simpleKey = 'cache:entry:Fixture:simple:' + self._arghash('alpha', 'beta')
    self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(simpleKey)))
    
    rockTag = pickle.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = pickle.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(16, len(rockTag))
    self.assertEqual(16, len(treeTag))
    
    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    
    self.testee.clean(('rock',))
    self.assertEqual(3, self.testee.backend.client.dbsize())
    
    self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(simpleKey)))
    
    self.assertIsNone(self.testee.backend.client.get('cache:tag:rock'))
    self.assertEqual(treeTag, pickle.loads(self.testee.backend.client.get('cache:tag:tree')))
    
    # stale still accessible, though only directly
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(3,            self.fixture.calls)
    self.assertEqual(5,            self.testee.backend.client.dbsize(), '+1 old tagged entry')
    
    self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(simpleKey)))
    
    self.assertNotEqual(rockTag, pickle.loads(self.testee.backend.client.get('cache:tag:rock')))
    rockTag = pickle.loads(self.testee.backend.client.get('cache:tag:rock'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(16, len(rockTag))
    self.assertEqual(treeTag, pickle.loads(self.testee.backend.client.get('cache:tag:tree')))
    
    # stale still accessible, though only directly
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    
    self.testee.clean(('rock', 'tree'))
    self.assertEqual(3, self.testee.backend.client.dbsize(), 'simaple, new tagged and old tagged')

    self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(simpleKey)))
    
    # new stale is accessible, though only directly
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(4,            self.fixture.calls)
    self.assertEqual(6,            self.testee.backend.client.dbsize(), '+2 old tagged entries')
    
    self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(simpleKey)))
    
    # new stale still accessible, though only directly
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    self.assertNotEqual(rockTag, pickle.loads(self.testee.backend.client.get('cache:tag:rock')))
    self.assertNotEqual(treeTag, pickle.loads(self.testee.backend.client.get('cache:tag:tree')))
    
    rockTag = pickle.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = pickle.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(16, len(rockTag))
    self.assertEqual(16, len(treeTag))
    
    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    
    self.testee.clean(('rock', 'tree'))
    self.testee.clean()
    self.assertEqual(0, self.testee.backend.client.dbsize())
  
  def testNested(self):
    self.assertEqual('beta+alpha', self.fixture.nested('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)
    key = 'cache:entry:Fixture:nested:' + self._arghash('alpha', 'beta')
    self.assertEqual('beta+alpha', pickle.loads(self.testee.backend.client.get(key)))
    key = 'cache:entry:Fixture:simple:' +  self._arghash('beta', 'alpha')
    self.assertEqual('ahpla+ateb', pickle.loads(self.testee.backend.client.get(key)))
    
  def testConcurrent(self):
    log = []
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)
    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      log.append(1)
      time.sleep(0.04)
      return '{0}-{1}'.format(a, b)[::2]
    
    threads = tuple(map(lambda i: threading.Thread(target = bar, args = ('alpha', 'beta')), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertEqual(1, sum(log))
    self.assertEqual(3, self.testee.backend.client.dbsize())
    
    aTag = pickle.loads(self.testee.backend.client.get('cache:tag:a'))
    zTag = pickle.loads(self.testee.backend.client.get('cache:tag:z'))
    self.assertFalse(aTag == zTag)
    self.assertEqual(16, len(aTag))
    self.assertEqual(16, len(zTag))
    
    tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
    key     = 'mk:alpha:beta:' + tagHash
    self.assertEqual('apabt', pickle.loads(self.testee.backend.client.get(key)))
    
    del log[:]
    self.testee.clean()
    self.testee.backend.lock = lambda k: hermes.backend.AbstractLock(k) # now see a dogpile 
    
    threads = tuple(map(lambda i: threading.Thread(target = bar, args = ('alpha', 'beta')), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertGreater(sum(log), 1, 'dogpile')
    # enries may be duplicated if tags overwrite
    self.assertGreaterEqual(self.testee.backend.client.dbsize(), 3)
    
    aTag = pickle.loads(self.testee.backend.client.get('cache:tag:a'))
    zTag = pickle.loads(self.testee.backend.client.get('cache:tag:z'))
    self.assertFalse(aTag == zTag)
    self.assertEqual(16, len(aTag))
    self.assertEqual(16, len(zTag))
    
    tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
    key     = 'mk:alpha:beta:' + tagHash
    self.assertEqual('apabt', pickle.loads(self.testee.backend.client.get(key)))


class TestRedisLock(test.TestCase):
  
  def setUp(self):
    cache = hermes.Hermes(hermes.backend.redis.Backend)
    cache.clean()
     
    self.testee = hermes.backend.redis.Lock('123', cache.backend.client)
  
  def testAcquire(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertFalse(self.testee.acquire(False))
        self.assertEqual('123', self.testee.key)
        self.assertEqual(900, self.testee.client.ttl(self.testee.key))
      finally:
        self.testee.release()

  def testRelease(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertFalse(self.testee.acquire(False))
        self.assertEqual('123', self.testee.key)
      finally:
        self.testee.release()
        self.assertIs(None, self.testee.client.get(self.testee.key))
    
  def testWith(self):
    with self.testee:
      self.assertFalse(self.testee.acquire(False))
      self.assertEqual('123', self.testee.key)
      
      client  = hermes.backend.redis.Backend(hermes.Mangler()).client
      another = hermes.backend.redis.Lock('234', client)
      with another:
        self.assertFalse(another.acquire(False))
        self.assertFalse(self.testee.acquire(False))
        self.assertEqual('234', another.key)
      
  def testConcurrent(self):
    log   = []
    check = threading.Lock()
    def target():
      with self.testee:
        log.append(check.acquire(False))
        time.sleep(0.05)
        check.release()
        time.sleep(0.05)
    
    threads = tuple(map(lambda i: threading.Thread(target = target), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertEqual([True] * 4, log)


class TestRedisPerformance(test.unittest.TestCase):
  
  def setUp(self):
    self.testee = hermes.Hermes(hermes.backend.redis.Backend, ttl = 360) 
    self.fixture = test.createFixture(self.testee)
    
  def tearDown(self):
    self.testee.clean()
  
  def testPerformance(self):
    test.benchmark(self.fixture)




'''
@author: saaj
'''


import threading
import time
import pickle
import telnetlib

import hermes.test as test
import hermes.backend.memcached


class TestMemcached(test.TestCase):
  
  def setUp(self):
    self.testee = hermes.Hermes(hermes.backend.memcached.Backend, ttl = 360) 
    self.fixture = test.createFixture(self.testee)
    
    self.testee.clean()
    
  def tearDown(self):
    self.testee.clean()
  
  def getSize(self):
    stats = self.testee.backend.client.get_stats()[0][1]
    try:
      # python3-memcached
      return int(stats[b'curr_items'])
    except KeyError:
      # pylibmc
      return int(stats['curr_items'])
  
  def testSimple(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.getSize())

    key = 'cache:entry:Fixture:simple:' + self._arghash('alpha', b = 'beta')
    for _ in range(4):
      self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', b = 'beta'))
      self.assertEqual(1, self.fixture.calls)
      self.assertEqual(1, self.getSize())
  
      self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(key)))
    
    self.fixture.simple.invalidate('alpha', b = 'beta')
    self.assertEqual(0, self.getSize())
    
    
    expectedPy2 = "])]'ammag'[(tes[+}]'ateb'[ :'ahpla'{"
    expectedPy3 = "]}'ammag'{[+}]'ateb'[ :'ahpla'{"
    key      = 'cache:entry:Fixture:simple:' + self._arghash({'alpha' : ['beta']}, [{'gamma'}])
    for _ in range(4):
      try: 
        self.assertEqual(expectedPy2, self.fixture.simple({'alpha' : ['beta']}, [{'gamma'}]))
      except AssertionError:
        self.assertEqual(expectedPy3, self.fixture.simple({'alpha' : ['beta']}, [{'gamma'}]))
        
      self.assertEqual(2, self.fixture.calls)
      self.assertEqual(1, self.getSize())
      
      try:
        self.assertEqual(expectedPy2, pickle.loads(self.testee.backend.client.get(key)))
      except AssertionError:
        self.assertEqual(expectedPy3, pickle.loads(self.testee.backend.client.get(key)))
    
  def testTagged(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.getSize())
    
    for _ in range(4):    
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', b = 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      self.assertEqual(3,       self.getSize())
      
      rockTag = pickle.loads(self.testee.backend.client.get('cache:tag:rock'))
      treeTag = pickle.loads(self.testee.backend.client.get('cache:tag:tree'))
      self.assertFalse(rockTag == treeTag)
      self.assertEqual(16, len(rockTag))
      self.assertEqual(16, len(treeTag))
      
      argHash = self._arghash('alpha', b = 'beta')
      tagHash = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
      key     = 'cache:entry:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
      self.assertEqual('ae-hl', pickle.loads(self.testee.backend.client.get(key)))
    
    self.fixture.tagged.invalidate('alpha', b = 'beta')
    
    self.assertEqual(2, self.getSize())
    
    rockTag = pickle.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = pickle.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertNotEqual(rockTag, treeTag)
    self.assertEqual(16, len(rockTag))
    self.assertEqual(16, len(treeTag))
    
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(3, self.fixture.calls)
      
      self.assertEqual(5, self.getSize())
      self.assertEqual(rockTag, pickle.loads(self.testee.backend.client.get('cache:tag:rock')))
      self.assertEqual(treeTag, pickle.loads(self.testee.backend.client.get('cache:tag:tree')))
      self.assertEqual(16, len(pickle.loads(self.testee.backend.client.get('cache:tag:ice'))))
      
    self.testee.clean(['rock'])
    
    self.assertEqual(4, self.getSize())
    self.assertIsNone(self.testee.backend.client.get('cache:tag:rock'))
    iceTag = pickle.loads(self.testee.backend.client.get('cache:tag:ice'))
    
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(5, self.fixture.calls)
      
      self.assertEqual(7,  self.getSize(), 'has new and old entries for tagged and tagged 2 + 3 tags')
      self.assertEqual(treeTag, pickle.loads(self.testee.backend.client.get('cache:tag:tree')))
      self.assertEqual(iceTag, pickle.loads(self.testee.backend.client.get('cache:tag:ice')))
      self.assertEqual(16, len(pickle.loads(self.testee.backend.client.get('cache:tag:rock'))))
      self.assertNotEqual(rockTag, pickle.loads(self.testee.backend.client.get('cache:tag:rock')))
    
  def testFunction(self):
    counter = dict(foo = 0, bar = 0)
    
    @self.testee
    def foo(a, b):
      counter['foo'] += 1
      return '{0}+{1}'.format(a, b)[::-1]
    
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)
    @self.testee(tags = ('a', 'z'), key = key, ttl = 1)
    def bar(a, b):
      counter['bar'] += 1
      return '{0}-{1}'.format(a, b)[::2]
    
    
    self.assertEqual(0, counter['foo'])
    self.assertEqual(0, self.getSize())
    
    key = 'cache:entry:foo:' + self._arghash('alpha', 'beta')
    for _ in range(4):
      self.assertEqual('ateb+ahpla', foo('alpha', 'beta'))
      
      self.assertEqual(1, counter['foo'])
      self.assertEqual(1, self.getSize())

      self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(key)))
    
    foo.invalidate('alpha', 'beta')
    self.assertIsNone(self.testee.backend.client.get(key))
    self.assertEqual(0, self.getSize())
    
    
    self.assertEqual(0, counter['bar'])
    self.assertEqual(0, self.getSize())
    
    for _ in range(4):
      self.assertEqual('apabt', bar('alpha', 'beta'))
      self.assertEqual(1,       counter['bar'])
      self.assertEqual(3,       self.getSize())
      
      aTag = pickle.loads(self.testee.backend.client.get('cache:tag:a'))
      zTag = pickle.loads(self.testee.backend.client.get('cache:tag:z'))
      self.assertFalse(aTag == zTag)
      self.assertEqual(16, len(aTag))
      self.assertEqual(16, len(zTag))
      
      tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
      key     = 'mk:alpha:beta:' + tagHash
      self.assertEqual('apabt', pickle.loads(self.testee.backend.client.get(key)))
    
    bar.invalidate('alpha', 'beta')
    self.assertEqual(1,  counter['foo'])
    self.assertIsNone(self.testee.backend.client.get(key))
    self.assertEqual(2, self.getSize())
    
    self.assertEqual('apabt', bar('alpha', 'beta'))
    self.assertEqual(2, counter['bar'])
    self.assertEqual(3, self.getSize())
    time.sleep(2)
    self.assertIsNone(self.testee.backend.client.get(key), 'should already expire')
    self.assertEqual(2, self.getSize())
    
    self.testee.clean(('a', 'z'))

  def testKey(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.getSize())
    
    for _ in range(4):
      self.assertEqual('apabt', self.fixture.key('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      self.assertEqual(3,       self.getSize())
      
      ashTag   = pickle.loads(self.testee.backend.client.get('cache:tag:ash'))
      stoneTag = pickle.loads(self.testee.backend.client.get('cache:tag:stone'))
      self.assertFalse(ashTag == stoneTag)
      self.assertEqual(16, len(ashTag))
      self.assertEqual(16, len(stoneTag))
      
      tagHash = self.testee.mangler.hashTags(dict(ash = ashTag, stone = stoneTag))
      key     = 'mykey:alpha:beta:' + tagHash
      self.assertEqual('apabt', pickle.loads(self.testee.backend.client.get(key)))
    
    self.fixture.key.invalidate('alpha', 'beta')
    
    self.assertIsNone(self.testee.backend.client.get(key))
    self.assertEqual(2, self.getSize())
    
    self.assertEqual(ashTag,   pickle.loads(self.testee.backend.client.get('cache:tag:ash')))
    self.assertEqual(stoneTag, pickle.loads(self.testee.backend.client.get('cache:tag:stone')))
      
    self.testee.clean(('a', 'z'))
    
  def testAll(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.getSize())
    for _ in range(4):    
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, self.fixture.all({'alpha' : 1}, ['beta']))
      self.assertEqual(1, self.fixture.calls)
      self.assertEqual(3, self.getSize())
      
      aTag = pickle.loads(self.testee.backend.client.get('cache:tag:a'))
      zTag = pickle.loads(self.testee.backend.client.get('cache:tag:z'))
      self.assertFalse(aTag == zTag)
      self.assertEqual(16, len(aTag))
      self.assertEqual(16, len(zTag))
      
      tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
      key = "mk:{'alpha':1}:['beta']:" + tagHash
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, pickle.loads(self.testee.backend.client.get(key)))
      
    self.fixture.all.invalidate({'alpha' : 1}, ['beta'])
    
    self.assertEqual(2, self.getSize())
    self.assertEqual(aTag, pickle.loads(self.testee.backend.client.get('cache:tag:a')))
    self.assertEqual(zTag, pickle.loads(self.testee.backend.client.get('cache:tag:z')))
    
    self.testee.clean(('a', 'z'))
  
  def testClean(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.getSize())
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    self.assertEqual(4,            self.getSize())
    
    self.testee.clean()
    
    self.assertEqual(2, self.fixture.calls)
    self.assertEqual(0, self.getSize())
    
  def testCleanTagged(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEqual(0, self.getSize())
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    self.assertEqual(4,            self.getSize())

    simpleKey = 'cache:entry:Fixture:simple:' + self._arghash('alpha', 'beta')
    self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(simpleKey)))
    
    rockTag = pickle.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = pickle.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(16, len(rockTag))
    self.assertEqual(16, len(treeTag))
    
    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    
    self.testee.clean(('rock',))
    self.assertEqual(3, self.getSize())
    
    self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(simpleKey)))
    
    self.assertIsNone(self.testee.backend.client.get('cache:tag:rock'))
    self.assertEqual(treeTag, pickle.loads(self.testee.backend.client.get('cache:tag:tree')))
    
    # stale still accessible, though only directly
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(3,            self.fixture.calls)
    self.assertEqual(5,            self.getSize(), '+1 old tagged entry')
    
    self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(simpleKey)))
    
    self.assertNotEqual(rockTag, pickle.loads(self.testee.backend.client.get('cache:tag:rock')))
    rockTag = pickle.loads(self.testee.backend.client.get('cache:tag:rock'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(16, len(rockTag))
    self.assertEqual(treeTag, pickle.loads(self.testee.backend.client.get('cache:tag:tree')))
    
    # stale still accessible, though only directly
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    
    self.testee.clean(('rock', 'tree'))
    self.assertEqual(3, self.getSize(), 'simaple, new tagged and old tagged')

    self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(simpleKey)))
    
    # new stale is accessible, though only directly
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    
    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(4,            self.fixture.calls)
    self.assertEqual(6,            self.getSize(), '+2 old tagged entries')
    
    self.assertEqual('ateb+ahpla', pickle.loads(self.testee.backend.client.get(simpleKey)))
    
    # new stale still accessible, though only directly
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    self.assertNotEqual(rockTag, pickle.loads(self.testee.backend.client.get('cache:tag:rock')))
    self.assertNotEqual(treeTag, pickle.loads(self.testee.backend.client.get('cache:tag:tree')))
    
    rockTag = pickle.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = pickle.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(16, len(rockTag))
    self.assertEqual(16, len(treeTag))
    
    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual('aldamg', pickle.loads(self.testee.backend.client.get(taggedKey)))
    
    
    self.testee.clean(('rock', 'tree'))
    self.testee.clean()
    self.assertEqual(0, self.getSize())
    
  def testNested(self):
    self.assertEqual('beta+alpha', self.fixture.nested('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)
    key = 'cache:entry:Fixture:nested:' + self._arghash('alpha', 'beta')
    self.assertEqual('beta+alpha', pickle.loads(self.testee.backend.client.get(key)))
    key = 'cache:entry:Fixture:simple:' + self._arghash('beta', 'alpha')
    self.assertEqual('ahpla+ateb', pickle.loads(self.testee.backend.client.get(key)))  
    
  def testConcurrent(self):
    log = []
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)
    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      log.append(1)
      time.sleep(0.04)
      return '{0}-{1}'.format(a, b)[::2]
    
    threads = tuple(map(lambda i: threading.Thread(target = bar, args = ('alpha', 'beta')), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertEqual(1, sum(log))
    self.assertEqual(3, self.getSize())
    
    aTag = pickle.loads(self.testee.backend.client.get('cache:tag:a'))
    zTag = pickle.loads(self.testee.backend.client.get('cache:tag:z'))
    self.assertFalse(aTag == zTag)
    self.assertEqual(16, len(aTag))
    self.assertEqual(16, len(zTag))
    
    tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
    key     = 'mk:alpha:beta:' + tagHash
    self.assertEqual('apabt', pickle.loads(self.testee.backend.client.get(key)))
    
    
    del log[:]
    self.testee.clean()
    self.testee.backend.lock = lambda k: hermes.backend.AbstractLock(k) # now see a dogpile 
    
    threads = tuple(map(lambda i: threading.Thread(target = bar, args = ('alpha', 'beta')), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertGreater(sum(log), 1, 'dogpile')
    # enries may be duplicated if tags overwrite
    self.assertGreaterEqual(self.getSize(), 3)
    
    aTag = pickle.loads(self.testee.backend.client.get('cache:tag:a'))
    zTag = pickle.loads(self.testee.backend.client.get('cache:tag:z'))
    self.assertFalse(aTag == zTag)
    self.assertEqual(16, len(aTag))
    self.assertEqual(16, len(zTag))

    
    self.testee.clean(('a', 'z'))


class TestMemcachedLock(test.TestCase):
  
  def setUp(self):
    cache = hermes.Hermes(hermes.backend.memcached.Backend)
    cache.clean()
    
    # make lock instance be created with client from ``hermes.Hermes`` instance at 
    # thread context to utilize ``hermes.backend.memcached.Backend`` ``thread.local``
    self.__class__.testee = property(lambda self: hermes.backend.memcached.Lock('123', cache.backend.client)) 
  
  def testAcquire(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertFalse(self.testee.acquire(False))
        self.assertEqual('123', self.testee.key)
      finally:
        self.testee.release()

  def testRelease(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertFalse(self.testee.acquire(False))
        self.assertEqual('123', self.testee.key)
      finally:
        self.testee.release()
    
  def testWith(self):
    with self.testee:
      self.assertFalse(self.testee.acquire(False))
      self.assertEqual('123', self.testee.key)
      
      client  = hermes.backend.memcached.Backend(hermes.Mangler()).client
      another = hermes.backend.memcached.Lock('234', client)
      with another:
        self.assertFalse(another.acquire(False))
        self.assertFalse(self.testee.acquire(False))
        self.assertEqual('234', another.key)
      
  def testConcurrent(self):
    log   = []
    check = threading.Lock()
    def target():
      with self.testee:
        log.append(check.acquire(False))
        time.sleep(0.05)
        check.release()
        time.sleep(0.05)
    
    threads = tuple(map(lambda i: threading.Thread(target = target), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))
    
    self.assertEqual([True] * 4, log)


class TestMemcachedPerformance(test.unittest.TestCase):
  
  def setUp(self):
    self.testee  = hermes.Hermes(hermes.backend.memcached.Backend, ttl = 360) 
    self.fixture = test.createFixture(self.testee)
    
  def tearDown(self):
    self.forceClean()
  
  def testPerformance(self):
    test.benchmark(self.fixture)
  
  def forceClean(self):
    '''What a poorly piece of software! Most of the issues that happen within the library
    developing are memcached issues. Most notable is that it won't flush damn keys. No matter
    how many times you call it to flush, or wait for clean state -- expired by flush call keys 
    are just stale with expiry timestamp in past. Current version is 1.4.14.
    '''
    
    telnet = telnetlib.Telnet('127.0.0.1', 11211)
    try:
      telnet.write(b'stats items\n')
      slablines = telnet.read_until(b'END').split(b'\r\n')
      keys      = set()
      for line in slablines:
        parts = line.decode().split(':')
        if len(parts) < 2:
          continue
        slab = parts[1]
        telnet.write('stats cachedump {0} 0\n'.format(slab).encode())
        cachelines = telnet.read_until(b'END').split(b'\r\n')
        for line in cachelines:
          parts = line.decode().split(' ', 3)
          if len(parts) < 2:
            continue
          keys.add(parts[1])
    finally:
      telnet.close()
    
    self.testee.backend.remove(keys)
    



'''
@author: saaj
'''


import time
import threading

try:
  import pylibmc as memcache
except ImportError:
  import memcache
  
from . import AbstractBackend, AbstractLock


__all__ = 'Lock', 'Backend'


class Lock(AbstractLock):
  '''Key-aware distrubuted lock'''
  
  client = None
  '''Memcached client'''
  
  timeout = 900
  '''Maximum TTL of lock, can be up to 30 days, 
  otherwise memcached will treated it as a unix timestamp of an exact date'''
  
  sleep = 0.1
  '''Amount of time to sleep per ``while True`` iteration when waiting'''
  
  
  def __init__(self, key, client, **kwargs):
    super(Lock, self).__init__(key)
    
    self.client = client
    
    self.sleep   = kwargs.get('lockSleep',   self.sleep)
    self.timeout = kwargs.get('lockTimeout', self.timeout)
    if self.timeout is None:
      self.timeout = 0

  def acquire(self, wait = True):
    while True:
      if self.client.add(self.key, 'locked', self.timeout):
        return True
      elif not wait:
        return False
      else:
        time.sleep(self.sleep)

  def release(self):
    self.client.delete(self.key)


class Backend(AbstractBackend):
  '''Memcached backend implementation'''
  
  _local = None
  '''Thread-local data'''
  
  _options = None
  '''Client options'''
  
  
  def __init__(self, mangler, **kwargs):
    self.mangler  = mangler
    self._options = kwargs
    self._local   = threading.local()
  
  @property
  def client(self):
    '''Thread-mapped memcached client accessor'''
    
    if not hasattr(self._local, 'client'):
      self._local.client = memcache.Client(self._options.get('servers', ['localhost:11211']))
      
    return self._local.client
  
  def lock(self, key):
    return Lock(self.mangler.nameLock(key), self.client, **self._options)
  
  def save(self, key = None, value = None, mapping = None, ttl = None):
    if not mapping:
      mapping = {key : value}
    mapping = {k : self.mangler.dumps(v) for k, v in mapping.items()}
    
    self.client.set_multi(mapping, ttl if ttl is not None else 0)
      
  def load(self, keys):
    if self._isScalar(keys):
      value = self.client.get(keys)
      if value is not None:
        value = self.mangler.loads(value)
      return value
    else:
      # python3 pylibmc returns byte keys
      return {
        k.decode() if isinstance(k, bytes) else k : self.mangler.loads(v) 
        for k, v in self.client.get_multi(tuple(keys)).items()
      }
  
  def remove(self, keys):
    if self._isScalar(keys):
      keys = (keys,)
      
    self.client.delete_multi(keys)

  def clean(self):
    self.client.flush_all()




# -*- coding: utf-8 -*-
'''
@author: saaj
'''


try:
  from setuptools import setup
except ImportError:
  from distutils.core import setup


setup(
  name             = 'HermesCache',
  version          = '0.5.2',
  author           = 'saaj',
  author_email     = 'mail@saaj.me',
  packages         = ['hermes', 'hermes.backend', 'hermes.test'],
  test_suite       = 'hermes.test',
  url              = 'https://bitbucket.org/saaj/hermes',
  license          = 'LGPL-2.1+',
  description      = 'Python caching library with tag-based invalidation and dogpile effect prevention',
  long_description = open('README.txt', 'rb').read().decode('utf-8'),
  platforms        = ['Any'],
  keywords         = 'python cache tagging redis memcached',
  classifiers      = [
    'Topic :: Software Development :: Libraries',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.3',    
    'Intended Audience :: Developers'
  ]
)




'''
@author: saaj
'''


import time
import importlib

redis = importlib.import_module('redis')

from . import AbstractBackend, AbstractLock


__all__ = 'Lock', 'Backend'


class Lock(AbstractLock):
  '''Key-aware distrubuted lock'''
  
  client = None
  '''Redis client'''
  
  timeout = 900
  '''Maximum TTL of lock'''
  
  sleep = 0.1
  '''Amount of time to sleep per ``while True`` iteration when waiting'''
  
  
  def __init__(self, key, client, **kwargs):
    super(Lock, self).__init__(key)
    
    self.client = client
    
    self.sleep   = kwargs.get('lockSleep',   self.sleep)
    self.timeout = kwargs.get('lockTimeout', self.timeout)
    if self.timeout is None:
      self.timeout = 0

  def acquire(self, wait = True):
    while True:
      if self.client.setnx(self.key, 'locked'):
        self.client.expire(self.key, self.timeout)
        return True
      elif not wait:
        return False
      else:
        time.sleep(self.sleep)

  def release(self):
    self.client.delete(self.key)


class Backend(AbstractBackend):
  '''Redis backend implementation'''
  
  client = None
  '''Redis client'''
  
  _options = None
  '''Lock options'''
  
  
  def __init__(self, mangler, **kwargs):
    super(Backend, self).__init__(mangler)
    
    self.client = redis.StrictRedis(**{
      'host'     : kwargs.pop('host',     'localhost'),
      'password' : kwargs.pop('password', None),
      'port'     : kwargs.pop('port',     6379),
      'db'       : kwargs.pop('db',       0)
    })
    
    self._options = kwargs
  
  def lock(self, key):
    return Lock(self.mangler.nameLock(key), self.client, **self._options)
  
  def save(self, key = None, value = None, mapping = None, ttl = None):
    if not mapping:
      mapping = {key : value}
    mapping = {k : self.mangler.dumps(v) for k, v in mapping.items()}
    
    if not ttl:
      self.client.mset(mapping)
    else:
      pipeline = self.client.pipeline()
      for k, v in mapping.items():
        pipeline.setex(k, ttl, v)
      pipeline.execute()
  
  def load(self, keys):
    if self._isScalar(keys):
      value = self.client.get(keys)
      if value is not None:
        value = self.mangler.loads(value)
      return value
    else:
      keys = tuple(keys) # py3 compat
      return {k : self.mangler.loads(v) for k, v in zip(keys, self.client.mget(keys)) if v is not None}
  
  def remove(self, keys):
    if self._isScalar(keys):
      keys = (keys,)
      
    self.client.delete(*keys)

  def clean(self):
    self.client.flushdb()

